/* 
 * Runset Init: Stub "init" utility (Linux specific)
 *
 * Copyright 1998 Michael Deutschmann
 *
 * This software package is subject to the GNU General Public License
 * version 2 or later (your choice) as published by the Free Software 
 * Foundation.
 *
 * THERE ARE NO WARRANTIES WHATSOEVER. 
 *
 */

#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/kdaemon.h>
#include <sys/reboot.h>
#include <sys/syslog.h>

#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* Process ID of the base runset */
static pid_t main_process;

/* Reboot mode
 *
 * 0 = halt system when runset exits
 * 1 = reboot system
 */
volatile int reboot_flag = 0;

/* Singleuser boot option
 *
 */
volatile int singleuser_flag = 0;

/* SIGCHLD handler
 *
 * If the base runset exited, shut down. We halt if we recieve a nonzero 
 * exit code.  
 *
 * As Process 1 we also must take responsibilty for orphan tasks. These are
 * logged, but are otherwise ignored.
 *
 */
static void
zombie (int signumber)
{
  pid_t process;
  int exit_status;

  process = waitpid (WAIT_ANY, &exit_status, WNOHANG);

  if (process < 1)
    return;

  if (process == main_process)
    {
      if (exit_status)
	syslog (LOG_EMERG, "Base process has failed. Halting.");
      else if (reboot_flag)
	syslog (LOG_NOTICE, "Base process exited. Standby for reboot.");
      else
	syslog (LOG_NOTICE, "Base process exited. Halting.");

      kill (-1, SIGSTOP);
      sync ();
      bdflush (1, 0);
      reboot (RB_ENABLE_CAD);
      reboot (reboot_flag ? RB_AUTOBOOT : RB_HALT_SYSTEM);
    }

  syslog (LOG_INFO, "orphan process (#%d) collected", process);

  zombie (0);			/* Tail recurse */
}

/* SIGTERM and SIGINT handler.
 * 
 * SIGTERM asks for halt, SIGINT asks for reboot. 
 *
 * This supports Linux's soft C-A-D, which sends SIGINT to process 1. 
 * Also, shutdown can be neatly done with `kill 1'.
 */
void
initiate_shutdown (int signumber)
{
  if (signumber == SIGINT)
    {
      reboot_flag = 1;
      syslog (LOG_NOTICE, "Recieved reboot signal. Initiating shutdown.");
    }
  else
    syslog (LOG_NOTICE, "Recieved halt signal. Initiating shutdown.");

  kill (main_process, SIGTERM);
}

/* The entry point.
 * 
 * We set up our signal handlers and Linux's soft reboot feature. 
 *
 * Then function as an update daemon (no sense wasting a seperate pid 
 * slot for that.)  
 *
 */
int
main (int argc, char **argv)
{
  struct sigaction sa;

  /* Idiot proofing... */
  if (getpid () != 1 || getuid () != 0)
    {
      printf ("This program is for booting only.\n");
      exit (EXIT_FAILURE);
    }

  /* Check if single-user mode was requested by LILO (or equivalent).
   */
  while (*argv)
    {
      if (!strcmp ("single", *argv))
	singleuser_flag = 1;
      argv++;
    }

  openlog ("INIT", LOG_CONS, LOG_CRON);

  syslog (LOG_INFO, "Starting...");

  sigemptyset (&sa.sa_mask);
  sa.sa_flags = SA_RESTART;

  sa.sa_handler = &zombie;
  sigaction (SIGCHLD, &sa, 0);

  main_process = fork ();

  if (main_process == -1)
    {
      syslog (LOG_EMERG, "cannot fork base process (%m)");
      reboot (RB_HALT_SYSTEM);
    };

  if (!main_process)
    {
      if (singleuser_flag)
	execl ("/etc/rs.d/init/singleuser", "singleuser", 0);
      else
	execl ("/etc/rs.d/init/multiuser", "multiuser", 0);
      exit (127);
    }

  sa.sa_handler = &initiate_shutdown;
  sigaction (SIGTERM, &sa, 0);
  sigaction (SIGINT, &sa, 0);
  reboot (RB_DISABLE_CAD);

  /* Function as bdflush updater daemon */

  while (1)
    {
      bdflush (1, 0);
      sleep (5);
    }
}
