/* 
 * Runset Init: Setup/Entry point
 *
 * Copyright 1998 Michael Deutschmann
 *
 * This software package is subject to the GNU General Public License
 * version 2 or later (your choice) as published by the Free Software 
 * Foundation.
 *
 * THERE ARE NO WARRANTIES WHATSOEVER. 
 *
 */

#define _GNU_SOURCE

#include "common.h"

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/syslog.h>


/* Planned shutdown flag.
 *
 * This flag is set by SIGINT or SIGTERM, and causes us to unwind all 
 * actions and exit.
 *
 * If there is an error, we will exit without setting this flag. If this 
 * happens, we return EXIT_FAILURE as our exit code.
 *
 */ 
volatile int planned_shutdown = 0;

/* Signal mask
 * 
 * BLOCKING is the set of signals we expect to recieve, and thus block
 * for most of the code. (It's not neccesarily the mask in effect during 
 * blocking.)
 *
 * NONBLOCKING is the normal signal mask. 
 */
sigset_t nonblocking, blocking;

/* Base on which the handler chain is built. */
static int
base (int shutdown)
{
  return 0;
}

/* Our signal handler. This handles all signals */
static void
signal_handler (int signum)
{
  switch (signum)
    {
    case SIGTERM:
    case SIGINT:
      planned_shutdown = 1;
      break;
#if 0
/* Kludge to enable some semblance of inetd functionality on broken Linux 
 * kernels (< 2.0.36) where SIGIO is sent prematurely for a listening socket. 
 */
    case SIGIO:
      setwake(time(0) + 2); 
      break;
#endif /* 0 */
    }
}

/* entry point
 * 
 * We check arguments, open the config file and syslog, and set the signals 
 * up. Then start the real code with next_command(). 
 */
int
main (int argc, char **argv)
{
  int x;
  struct sigaction sa;

  static struct option lopts[] =
  {
    {"version", no_argument, 0, 256},
    {"help", no_argument, 0, 257}
  };

  static const char *Usage = "runset <config_file>\n";
  static const char *Version = PACKAGE " " VERSION "\n";

  while ((x = getopt_long (argc, argv, "", lopts, 0)) != -1)
    {
      switch (x)
	{
	case 256:
	  printf (Version);
	  return EXIT_FAILURE;
	case 257:
	  printf (Usage);
	  return EXIT_FAILURE;
	}
    }

  if (argc != 2)
    {
      fprintf (stderr, Usage);
      return EXIT_FAILURE;
    }

  openlog (program_invocation_short_name, LOG_PID | LOG_CONS,LOG_CRON);

  config_file = fopen (argv[1], "r");

  {
    /* prevent config file being passed to children */
    int fd;

    fd = fileno (config_file);
    fcntl (fd, F_SETFD, fcntl (fd, F_GETFD, 0) | FD_CLOEXEC);
  }

  if (!config_file)
    {
      syslog (LOG_ERR, "can't load config file `%s' (%m)", argv[1]);
      return EXIT_FAILURE;
    }

  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = &signal_handler;

  sigaction (SIGIO, &sa, 0);
  sigaction (SIGCHLD, &sa, 0);
  sigaction (SIGALRM, &sa, 0);
  sigaction (SIGTERM, &sa, 0);
  sigaction (SIGINT, &sa, 0);

  sigemptyset (&blocking);
  sigaddset (&blocking, SIGIO);
  sigaddset (&blocking, SIGCHLD);
  sigaddset (&blocking, SIGALRM);
  sigaddset (&blocking, SIGTERM);
  sigaddset (&blocking, SIGINT);
  sigprocmask (SIG_BLOCK, &blocking, &nonblocking);

  next_command (&base);
  return planned_shutdown ? EXIT_SUCCESS : EXIT_FAILURE;
}
