/* 
 * Runset Init: Commands - simple process execution
 *
 * Copyright 1998 Michael Deutschmann
 *
 * This software package is subject to the GNU General Public License
 * version 2 or later (your choice) as published by the Free Software 
 * Foundation.
 *
 * THERE ARE NO WARRANTIES WHATSOEVER. 
 *
 */

#define _GNU_SOURCE

#include "common.h"
#include "dispatch.h"

#include <time.h>
#include <assert.h>
#include <unistd.h>
#include <signal.h>

/* Launch a process, then go on. We don't care when/if it exits. */
void
act_once (hchain top, cvec vector, cenv env)
{
  time_t grace = 0;
  int grace_length;

  int handler (int shutdown)
  {
    assert (!(*top) (0));

    if (shutdown)
      return graceful_shutdown (top, grace_length, &grace);

    return 0;
  }

  if (verify_parameters (vector, 1, 0))
      return;

  grace_length = get_env_number (env, "grace", 10);

  if (!launch (top, vector + 1, env, STDIN_FILENO));
  next_command (&handler);
}

/* Launch a process. If it fails before we are told to shutdown, shut 
 * down the entire runset
 */
void
act_watch (hchain top, cvec vector, cenv env)
{
  time_t grace = 0;
  int grace_length;

  int handler (int shutdown)
  {
    assert (!(*top) (0));

    if (shutdown)
      return graceful_shutdown (top, grace_length, &grace);

    if (!count (top))		/* if process died */
      raise (SIGINT);		/* then kill ourself */

    return 0;
  }

  if (verify_parameters (vector, 1, 0))
      return;

  grace_length = get_env_number (env, "grace", 10);

  if (!launch (top, vector + 1, env, STDIN_FILENO));
  next_command (&handler);
}

/* Lauch a process, wait for it to finish, then go on */
void
act_wait (hchain top, cvec vector, cenv env)
{
  time_t grace = 0;
  int grace_length;

  int handler (int shutdown)
  {
    assert (!(*top) (0));

    if (shutdown)
      return graceful_shutdown (top, grace_length, &grace);

    return count (top);
  }

  if (verify_parameters (vector, 1, 0))
      return;

  grace_length = get_env_number (env, "grace", 10);

  if (!launch (top, vector + 1, env, STDIN_FILENO))
    next_command (&handler);
}

/* Don't do anything at startup, but launch process just before shutting
 * down */
void
act_leave (hchain top, cvec vector, cenv env)
{
  int launched = 0;
  int handler (int shutdown)
  {
    assert (!(*top) (0));

    if (shutdown)
      {
	if (launched)
	  return count (top);
	else
	  {
	    launch (top, vector + 1, env, STDIN_FILENO);
	    launched = 1;
	    return 1;
	  }
      }
    return 0;
  }

  if (verify_parameters (vector, 1, 0))
    return;

  next_command (&handler);
}

/* Launch a process again&again if necessary. */
void
act_respawn (hchain top, cvec vector, cenv env)
{
  time_t grace = 0, last_launch = 0;
  int grace_length, quiet_time, max_launches, launches = 0;

  int handler (int shutdown)
  {
    assert (!(*top) (0));

    if (shutdown)
      return graceful_shutdown (top, grace_length, &grace);

    if ((last_launch + quiet_time) <= now)
      launches = 0;

    if (launches >= max_launches)
      {
	setwake (last_launch + quiet_time);
	return 0;
      }

    if (!count (top))
      {
	last_launch = now;
	launches++;

	launch (top, vector + 1, env, STDIN_FILENO);
      }
    return 0;
  }

  if (verify_parameters (vector, 1, 0))
    return;

  grace_length = get_env_number (env, "grace", 10);
  quiet_time = get_env_number (env, "quiet", 60);
  max_launches = get_env_number (env, "max", 3);

  next_command (&handler);
}
