/* 
 * Runset Init: Switch daemon control module
 *
 * Copyright 1998 Michael Deutschmann
 *
 * This software package is subject to the GNU General Public License
 * version 2 or later (your choice) as published by the Free Software 
 * Foundation.
 *
 * THERE ARE NO WARRANTIES WHATSOEVER. 
 *
 */

#define _GNU_SOURCE

#include "common.h"

#include <sys/socket.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/syslog.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <fcntl.h>

/* signal sets used to prevent an update/reap while a command is processed.
 * 
 * NONBLOCKING is exposed because switcher.c needs to clear the signal mask
 * before launching a child process.
 *
 */
sigset_t nonblocking;
static sigset_t blocking;

/* Flag to detect termination signals */
volatile int keep_going = 1;

/* SIGTERM/SIGINT handler 
 * 
 * Simply clears a keepalive flag.
 */
static void
term_handler (int signum)
{
  keep_going = 0;
}

/* SIGCHLD handler
 * 
 * Pass along information about exited processes to the switcher.c module,
 * then do an update().
 *
 */
static void
zombie_handler (int signum)
{
  pid_t process;

  process = waitpid (WAIT_ANY, 0, WNOHANG);

  if (process > 0)
    {
      reap (process);
      zombie_handler (0);
    }
  else
    {
      update ();
    }
}

/* React to a command from `request'.
 * 
 * The protocol is
 *  '+RUNSET,UID\n' add a request for RUNSET by user UID
 *  '-RUNSET,UID\n' remove a request " " " " "
 *  '!RUNSET\n' cancel all requests for RUNSET - force it to shut down.
 *  '?RUNSET\n' list all users who have requests for RUNSET.
 *            (this is done by the all the above commands)
 *  '*\n' Dump a list of all active runsets.
 *
 * (No promise is made to keep the protocol constant between revisions)
 *
 * The reply is handled in switcher.c, and consists of
 *  "RUNSET: UID_1 UID_2 UID_3 ... UID_N\n"
 * 
 * Where RUNSET is the runset mentioned and UID_? is users requesting it.
 *
 */
static void
accept_command (int fd)
{
  FILE *stream;
  char *command;
  size_t cs;

  stream = fdopen (fd, "r+");

  if (!stream)
    {
      syslog (LOG_WARNING, "Couldn't open high-level stream (%m)");
      close (fd);
      return;
    }

  command = 0;
  cs = 0;

  if (getline (&command, &cs, stream))
    {
      char *runset;
      char *user;
      uid_t userid;

      runset = command + 1;
      user = strpbrk (command, ",");
      if (user)
	{
	  *user = 0;
	  user++;
	  userid = strtol (user, 0, 0);
	}
      else
	{
	  runset[strlen (runset) - 1] = 0;	/* Clip off newline */
	  userid = 0;
	}

      /* Protocol
       *  +<runset>,<user>   Request
       *  -<runset>,<user>   Remove request
       *  !<runset>          Force all requests dropped
       *  ?<runset>          Dump status of given runset.
       *  *                  Dump status of all runsets.
       */

      switch (command[0])
	{
	case '+':
	  syslog (LOG_INFO,
		  "Request for `%s' added for user %d", runset, userid);
	  add_request (runset, userid);
	  dump_requests (stream, runset);
	  break;
	case '-':
	  syslog (LOG_INFO,
		  "Request for `%s' removed for user %d", runset, userid);
	  remove_request (runset, userid);
	  dump_requests (stream, runset);
	  break;
	case '!':
	  syslog (LOG_INFO, "All requests for `%s' removed", runset, userid);
	  cancel_requests (runset);
	  dump_requests (stream, runset);
	  break;
	case '?':
	  dump_requests (stream, runset);
	  break;
	case '*':
	  dump_all (stream);
	  break;
	case 0:
	  syslog (LOG_WARNING,
		  "Recieved empty command",
		  command[0]);
	  break;
	default:
	  syslog (LOG_WARNING,
		  "Recieved unknown command ('%c')",
		  command[0]);
	}
      free (command);
    }
  else
    {
      if (feof (stream))
	syslog (LOG_WARNING, "EOF recieving command");
      else
	syslog (LOG_WARNING, "Error recieving command (%m)");
    }

  fclose (stream);
}

/* Entry point
 * 
 * Set up signal handlers, then listen on a control socket (/dev/switch) for
 * commands.
 */
int
main (int argc, char **argv)
{
  struct sigaction sa;
  struct sockaddr_un address =
  {AF_UNIX, "/dev/switch"};
  int lsocket;

  {
    static const struct option lopts[] =
    {
      {"version", no_argument, 0, 256},
      {"help", no_argument, 0, 257}
    };

    static const char *Usage = "switch\n";
    static const char *Version = "switch RUNSETS... (" PACKAGE ") " VERSION "\n";

    int x;

    while ((x = getopt_long (argc, argv, "", lopts, 0)) != -1)
      switch (x)
	{
	case 256:
	  printf (Version);
	  return EXIT_FAILURE;
	case 257:
	  printf (Usage);
	  return EXIT_FAILURE;
	}

  }

  sa.sa_handler = &zombie_handler;
  sa.sa_flags = 0;		/* No SA_RESTART - we need to interrupt accept() */
  sigemptyset (&sa.sa_mask);
  sigaction (SIGCHLD, &sa, 0);

  sa.sa_handler = &term_handler;
  sigaction (SIGTERM, &sa, 0);
  sigaction (SIGINT, &sa, 0);

  sigemptyset (&blocking);
  sigaddset (&blocking, SIGCHLD);

  openlog (program_invocation_short_name, LOG_CONS, LOG_CRON);

  unlink (address.sun_path);

  lsocket = socket (PF_UNIX, SOCK_STREAM, 0);
  if (lsocket == -1)
    {
      syslog (LOG_ERR, "Can't open socket (%m)");
      exit (EXIT_FAILURE);
    }

  if (bind (lsocket, &address, sizeof (address)))
    {
      syslog (LOG_ERR, "Can't bind socket to /dev/switch (%m)");
      exit (EXIT_FAILURE);
    }

  fcntl (lsocket, F_SETFD, fcntl (lsocket, F_GETFD) | FD_CLOEXEC);

  if (listen (lsocket, 0))
    {
      syslog (LOG_ERR, "Can't listen (%m)");
      exit (EXIT_FAILURE);
    }

  /* Preload active runsets... */
  while (argc > 1)
    add_request (argv[--argc], 0);

  raise (SIGCHLD);

  while (keep_going)
    {
      int asocket;

      asocket = accept (lsocket, 0, 0);
      if (asocket > 0)
	{
	  sigprocmask (SIG_BLOCK, &blocking, &nonblocking);
	  accept_command (asocket);
	  raise (SIGCHLD);
	  sigprocmask (SIG_SETMASK, &nonblocking, 0);
	}
      else if (errno != EINTR)
	syslog (LOG_WARNING, "Error on accepting connection (%m)");
    }

  /* Remove SIGCHLD handler. For the deathwatch, we run synchronously */
  sa.sa_handler = SIG_DFL;
  sigaction (SIGCHLD, &sa, 0);

  cancel_all ();

  close (lsocket);
  unlink ("/dev/switch");

  while (update ())
    reap (waitpid (WAIT_ANY, 0, 0));

  return EXIT_SUCCESS;
}
